From cd78dcdf01890ad69d962ef18feef79ae2c4cf2a Mon Sep 17 00:00:00 2001
From: "Robert A. Burks" <rburksdev@gmail.com>
Date: Wed, 6 Mar 2024 18:56:31 -0500
Subject: [PATCH 02/31] Fix multiple watcher calls in set_default_internal
 (bug#00001)

* src/data.c (set_default_internal):
Variable watchers will now execute one time if called using an alias.
(bug#00001)
---
 src/data.c | 43 ++++++++++++++++++++++++++-----------------
 1 file changed, 26 insertions(+), 17 deletions(-)

diff --git a/src/data.c b/src/data.c
index c42424497ad..48da5cee429 100644
--- a/src/data.c
+++ b/src/data.c
@@ -1986,31 +1986,21 @@ set_default_internal (Lisp_Object symbol, Lisp_Object value,
 {
   CHECK_SYMBOL (symbol);
   struct Lisp_Symbol *sym = XSYMBOL (symbol);
-  switch (sym->u.s.trapped_write)
+  /* Update for aliasing and sanitize the input to ensure only a bare
+     symbol is stored.  It may be accessed with positions disabled.  */
+ start:
+  XSETSYMBOL (symbol, sym);
+
+  if (sym->u.s.trapped_write == SYMBOL_NOWRITE)
     {
-    case SYMBOL_NOWRITE:
       if (NILP (Fkeywordp (symbol))
           || !EQ (value, Fsymbol_value (symbol)))
         xsignal1 (Qsetting_constant, symbol);
       else
         /* Allow setting keywords to their own value.  */
-        return;
-
-    case SYMBOL_TRAPPED_WRITE:
-      /* Don't notify here if we're going to call Fset anyway.  */
-      if (sym->u.s.redirect != SYMBOL_PLAINVAL
-          /* Setting due to thread switching doesn't count.  */
-          && bindflag != SET_INTERNAL_THREAD_SWITCH)
-        notify_variable_watchers (symbol, value, Qset_default, Qnil);
-      break;
-
-    case SYMBOL_UNTRAPPED_WRITE:
-      break;
-
-    default: emacs_abort ();
+	return;
     }
 
- start:
   switch (sym->u.s.redirect)
     {
     case SYMBOL_VARALIAS: sym = SYMBOL_ALIAS (sym); goto start;
@@ -2019,6 +2009,11 @@ set_default_internal (Lisp_Object symbol, Lisp_Object value,
       {
 	struct Lisp_Buffer_Local_Value *blv = SYMBOL_BLV (sym);
 
+	/* If watched notify unless thread switching */
+	if (sym->u.s.trapped_write == SYMBOL_TRAPPED_WRITE
+	    && bindflag != SET_INTERNAL_THREAD_SWITCH)
+	  notify_variable_watchers (symbol, value, Qset_default, Qnil);
+
 	/* Store new value into the DEFAULT-VALUE slot.  */
 	XSETCDR (blv->defcell, value);
 
@@ -2039,6 +2034,20 @@ set_default_internal (Lisp_Object symbol, Lisp_Object value,
 	    int offset = XBUFFER_OBJFWD (valcontents)->offset;
 	    int idx = PER_BUFFER_IDX (offset);
 
+	    /* If watched notify unless thread switching */
+	    if (sym->u.s.trapped_write == SYMBOL_TRAPPED_WRITE
+		&& bindflag != SET_INTERNAL_THREAD_SWITCH)
+	      {
+		/* If a let shadows buffer binding we can reach here
+		   from a 'makunbound' within a let.  */
+		bool voide = BASE_EQ (value, Qunbound);
+
+		notify_variable_watchers (symbol, voide? Qnil : value,
+					  (bindflag == SET_INTERNAL_BIND? Qlet :
+					   bindflag == SET_INTERNAL_UNBIND? Qunlet :
+					   voide? Qmakunbound : Qset_default), Qnil);
+	      }
+
 	    set_per_buffer_default (offset, value);
 
 	    /* If this variable is not always local in all buffers,
-- 
2.34.1

